#include <errno.h>
#include <fcntl.h>
#include <netinet/in.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/ioctl.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <sys/soundcard.h>
#include <sys/time.h>
#include <unistd.h>

typedef struct
{
    short l, r;
} stereoent;

struct task
{
    int fd, Bits, Stereo, Rate;
    struct task *next;
};

static const unsigned short ErecPort = 6130;

/* returns -1 if socket was closed. */
static int senddata(struct task *t, const void *p, unsigned c)
{
    const unsigned char *q = p;
    unsigned sent=0;
    while(c)
    {
        unsigned try = (sent&1) ? 1 : c;
        int n;
        n = send(t->fd, q, try, MSG_NOSIGNAL);
        if(n < 0 && errno != EAGAIN)return -1;
        if(n < 0)n = 0;
        sent += n;
        if(!(sent & 1))
        {
            struct timeval tv = {0, 100000};
            fd_set fds;
            FD_ZERO(&fds);
            FD_SET(t->fd, &fds);
            if(select(t->fd+1, NULL, &fds, NULL, &tv) != 1)
            {
                /* Don't keep other streams waiting */
                return 0;
            }
        }
        q += n;
        c -= n;
    }
    return 0;
}
static int sendtask(struct task *t, stereoent *buf, unsigned count, const int Rate)
{
    const double suhde = (double)t->Rate / Rate;
    const double dresultlen = count * suhde;
    const unsigned resultlen = dresultlen;
    
    stereoent dest[resultlen];
    
    double step = (double)count / resultlen;
    double pos = 0;
    unsigned x;
    for(x=0; x<resultlen; ++x)
    {
        dest[x] = buf[(unsigned)pos];
        pos += step;
    }
    if(t->Bits == 16)
    {
        if(t->Stereo)
        {
            return senddata(t, &dest, sizeof dest);
        }
        else
        {
            short tmp[resultlen];
            for(x=0; x<resultlen; ++x)
                tmp[x] = (dest[x].l + dest[x].r) / 2;
            return senddata(t, &tmp, sizeof tmp);
        }
    }
    else if(t->Bits == 8)
    {
        if(t->Stereo)
        {
            unsigned char tmp[resultlen*2];
            unsigned p=0;
            for(x=0; x<resultlen; ++x)
            {
                tmp[p++] = 128 + dest[x].l/256;
                tmp[p++] = 128 + dest[x].r/256;
            }
            return senddata(t, &tmp, sizeof tmp);
        }
        else
        {
            unsigned char tmp[resultlen];
            for(x=0; x<resultlen; ++x)
                tmp[x] = 128 + (dest[x].l + dest[x].r) / 512;
            return senddata(t, &tmp, sizeof tmp);
        }
    }
    return -1;
}

/* returns the number of entries read */
static unsigned getstereoents(int fd, stereoent *buf, unsigned maxnum)
{
    const unsigned entsize = sizeof(*buf);
    static unsigned char orphanbuf[sizeof(*buf)];
    static unsigned orphan = 0;
    unsigned char *dest = (unsigned char *)buf;
    unsigned maxbyte = maxnum * entsize;
    unsigned result, bytenum;
    if(orphan)
    {
        memcpy(dest, orphanbuf, orphan);
        dest += orphan;
        maxbyte -= orphan;
    }
    bytenum = read(fd, dest, maxbyte) + orphan;
    dest -= orphan;
    orphan = bytenum % entsize;
    result = bytenum / entsize;
    if(orphan)
    {
        bytenum -= orphan;
        memcpy(orphanbuf, &dest[bytenum], orphan);
    }
    return result;
}

static void Serveri(int fd)
{
    struct task *tasks = NULL;
    struct sockaddr_in madr;
    const int Bits = 16;
    const int Stereo = 1;
    const int Rate = 44100;
    unsigned radrsize = sizeof madr;
    int Plug;
    
    if(ioctl(fd, SNDCTL_DSP_SAMPLESIZE, &Bits) < 0
    || ioctl(fd, SNDCTL_DSP_STEREO,     &Stereo) < 0
    || ioctl(fd, SNDCTL_DSP_SPEED,      &Rate) < 0)
    {
        exit(-1);
    }
    
    fclose(stdin);
    fclose(stdout);
    fclose(stderr);
    chdir("/");
    
    if(fork() > 0)_exit(0);
    setsid();
    
    memset(&madr, 0, sizeof madr);
    madr.sin_port = htons(ErecPort);
    madr.sin_family = AF_INET;
    madr.sin_addr.s_addr = 0; /* any */
    Plug = socket(madr.sin_family, SOCK_STREAM, 0);
    { int dummy;
    setsockopt(Plug, SOL_SOCKET, SO_REUSEADDR, &dummy, sizeof dummy); }
    bind(Plug, (struct sockaddr *)&madr, sizeof(struct sockaddr));
    listen(Plug, 10);
    fcntl(Plug, F_SETFL, fcntl(Plug, F_GETFL) | O_NONBLOCK);
    fcntl(fd,   F_SETFL, fcntl(fd,   F_GETFL) | O_NONBLOCK);
    for(;;)
    {
        struct sockaddr_in radr;
        stereoent buf[512];
        unsigned entnum;
        int Sock=0;
        struct task **tmp;
        fd_set fds;
        struct timeval tv = {0, 100000};
        FD_ZERO(&fds);
        FD_SET(Plug, &fds); if(Plug>=Sock)Sock = Plug+1;
        FD_SET(fd,   &fds); if(fd  >=Sock)Sock = fd+1;
        
        select(Sock, &fds, NULL, NULL, &tv);
        
        if(FD_ISSET(Plug, &fds))
        {
            int Sock = accept(Plug, (struct sockaddr *)&radr, &radrsize);
            if(Sock >= 0)
            {
                unsigned char c;
                struct task *tmp = (struct task *)malloc(sizeof(*tmp));
                tmp->fd = Sock;
                read(Sock, &c, 1); tmp->Bits = c;
                read(Sock, &c, 1); tmp->Stereo = c;
                read(Sock, &c, 1); tmp->Rate = c;
                read(Sock, &c, 1); tmp->Rate |= c<<8;
                fcntl(Sock, F_SETFL, fcntl(Sock, F_GETFL) | O_NONBLOCK);
                tmp->next = tasks;
                tasks = tmp;
            }
        }
        if(FD_ISSET(fd, &fds))
        {
            int c=0;
            entnum = getstereoents(fd, buf, (sizeof(buf) / sizeof(buf[0])));
            for(tmp = &tasks; *tmp; )
            {
                if(sendtask(*tmp, buf, entnum, Rate) < 0)
                {
                    struct task *t = (*tmp)->next;
                    close((*tmp)->fd);
                    free(*tmp);
                    *tmp = t;
                    ++c;
                }
                else
                    tmp = &(*tmp)->next;
            }
            if(c && !tasks)
            {
                /* All clients are gone. I'm dead.
                 * In a typical case, nobody even
                 * realizes that this was process
                 * here... Unless they don't have
                 * TCP/IP, which is exceptional.
                 */
                _exit(0);
            }
        }
    }
}

static int ServeriYhteys(int Bits, int Stereo, int Rate)
{
    struct sockaddr_in madr;
    char c;
    int sock;
    int fd = open("/dev/dsp", O_RDONLY);
    int err;
    if(fd >= 0) { int p = fork(); if(!p)Serveri(fd); close(fd); sleep(1); }
    err = errno;
    
    sock = socket(AF_INET, SOCK_STREAM, 0);
    if(sock < 0) { perror("socket"); return -1; }
    memset(&madr, 0, sizeof madr);
    madr.sin_port   = htons(ErecPort);
    madr.sin_family = AF_INET;
    madr.sin_addr.s_addr = 0x0100007F; /* localhost */
    if(connect(sock, (struct sockaddr *)&madr, sizeof madr) < 0)
    {
        if(err)
            fprintf(stderr, "/dev/dsp open failed (%s), and erec is not already running.\n", strerror(err));
        close(sock);
        return -1;
    }
    c = Bits;     send(sock, &c, 1, MSG_NOSIGNAL);
    c = Stereo;   send(sock, &c, 1, MSG_NOSIGNAL);
    c = Rate&255; send(sock, &c, 1, MSG_NOSIGNAL);
    c = Rate>>8;  send(sock, &c, 1, MSG_NOSIGNAL);
    return sock;
}

#include "argh.h"

struct Option Options[] =
{
    {'2', '2', 0, "stereo",  "Specifies stereo sound. Default is mono."},
    {'m', 'm', 0, "mono",    "Redundant. It's here for esd compatibility."},
    {'b', 'b', 0, "byte",    "8-bit mode. Normally 16-bit mode."},
    {'6', '6', 0, "word",    "16-bit mode. Redundant."},
    {'r', 'r', 0, "rate",    "Specifies recording rate. 43200 is default."},
    {'?', 'h', 0, "help",    "Help"},
    {'V', 'V', 0, "version", "Version information"},
    {0,0,0,0,0}
};

const char ProgName[] = "erec";

static int stereo=0;
static int Bits  =16;
static int Rate  =43200;
static const char *host = NULL;

void DefArg(const char *arg)
{
    if(host)ArgError(2, arg);
    host = arg;
}
void Case(int id, const char **dummy)
{
    switch(id)
    {
        case 'V':
            PrintVersionInfo();
            exit(EXIT_SUCCESS);
        case '?':
            Heelp = 1;
            break;
          case '2':
              ++stereo;
              break;
          case 'b':
              Bits = 8;
              break;
          case '6':
              Bits = 16;
              break;
          case 'm':
              stereo = 0;
              break;
          case 'r':
             Rate = (int)ParamRange(100,65535, dummy);
             break;
        case -1:
            /* Finish args, set default values if necessary */
            if(!host)host = "10.104.2.2";             
    }
}
void PrintVersionInfo(void)
{
    printf("%s - polydriving recording server v"VERSION" (C) 1992,2000 Bisqwit\n", ProgName);
}
void PrintDescription(void)
{
    printf(
        "\nAllows multiple applications request recorded data"
        "\nat the same time with different stream attributes.\n");
}

#include "cat.h"

int main(int argc, const char *const *argv)
{
    int fd, saved_stderr;
    
    ReadArgs(argc, argv,
        " [<host> | none]",
        "If host is other than none, the output will be sent with esdcat.\n");
    
    fd = ServeriYhteys(Bits, stereo, Rate);
    if(fd < 0)
    {
        perror(*argv);
        return errno;
    }
    
    dup2(fd, 0);
    
    saved_stderr = dup(2);
    
    close(2);
    if(!strcmp(host, "none"))
        cat(); /* doesn't return */
    else
    {
        char Buf[256];
        close(1);
        sprintf(Buf, "%d", Rate);
        execlp("esdcat", "esdcat", "-r", Buf, "-s", host, stereo?NULL:"-m", NULL);
    }
    
    dup2(saved_stderr, 2);
    perror("execlp");

    return -1;
}
